#include "polyhedron.h"
#include "randomizer.h" 
#include <gl\gl.h>
#include <gl\glu.h>
#include <gl\glut.h>
#include <assert.h>

//#define TESTRADIUSACCURACY
/* If I switch on TESTRADIUSACCURACY, I draw a wireframe sphere of _radius 
_radius at the same spot as each polyhedron to check of my _radius has the 
right value. */

//Helper methods 

/* glutWireCylinder isn't defined, so we define it. */

GLUquadricObj *pquadObj = gluNewQuadric();

void glutWireCylinder(double bottomradius, double topradius, double height,
	int slices, int stacks)
{
	gluQuadricDrawStyle(pquadObj, GLU_LINE);
	gluQuadricNormals(pquadObj, GLU_SMOOTH);
	gluCylinder(pquadObj, bottomradius, topradius, height, slices, stacks);
}

void glutSolidCylinder(double bottomradius, double topradius, double height,
	int slices, int stacks)
{
	gluQuadricDrawStyle(pquadObj, GLU_FILL);
	gluQuadricNormals(pquadObj, GLU_SMOOTH);
	gluCylinder(pquadObj, bottomradius, topradius, height, slices, stacks);
}

void setMaterialColor(cVector3 color)
{
  	float r = (float)color.x();
	float g = (float)color.y();
	float b = (float)color.z();
	CLAMP(r, 0.0, 1.0); //Use a macro from realnumber.h
	CLAMP(g, 0.0, 1.0);
	CLAMP(b, 0.0, 1.0);
	GLfloat materialAmbDiff[4]={ r, g, b, 1.0f };
	::glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, materialAmbDiff );
}

//******** CONST STATICS **********************************************
const int cPolyhedron::TETRAHEDRON = 0;
const int cPolyhedron::CUBE = 1;
const int cPolyhedron::OCTAHEDRON = 2;
const int cPolyhedron::DODECAHEDRON = 3;
const int cPolyhedron::ICOSOHEDRON = 4;
const int cPolyhedron::SPHERE = 5;
const int cPolyhedron::CYLINDER = 6;
const int cPolyhedron::CONE = 7;
const int cPolyhedron::TORUS = 8;
const int cPolyhedron::TEAPOT = 9;

//------------Constructors, Destructor, and Copiers---------------

cPolyhedron::cPolyhedron(int gluttype, Real radius,
	int slices, int stacks, Real edge,
	Real tuberadius, Real bottomradius,
	Real topradius, Real height):
		_gluttype(gluttype), _actualradius(radius),
		_radius(radius),
		_slices(slices), _stacks(stacks), _edge(edge),
		_tuberadius(tuberadius), _bottomradius(bottomradius),
		_topradius(topradius), _height(height),
		_color(cVector(1.0, 0.0, 0.0)), _alpha(1.0),
		_filled(true), _linewidth(2.0),
		_offset(0.0,0.0,0.0), _position(0,0,0),
		_spinangle(0.0), _scale(1.0, 1.0, 1.0)
{
	fixActualRadius();
	_spinaxis = cVector::randomUnitVector(); 
	_spinrate = cRandomizer::pinst()->randomReal(3.0*(2*PI), 20.0*(2*PI));
		//Between 3 and 20 full rotations per second. 
} 

void cPolyhedron::fixActualRadius()
{
	Real bigradius; //Need bigradius for CYLINDER
	//Need alpha, bega, dodecradius for DODECAHEDRON
	static Real  alpha = sqrt(2.0 / (3.0 + sqrt(5.0)));
	static Real beta = 1.0 + sqrt(6.0 / (3.0 + sqrt(5.0)) -
		2.0 + 2.0 * sqrt(2.0 / (3.0 + sqrt(5.0))));
	static Real dodecradius = sqrt(alpha*alpha + beta*beta);
	_offset = cVector(0.0, 0.0, 0.0);

	switch(_gluttype)
	{
		case TETRAHEDRON:
			break;
		case CUBE:
			_actualradius = maxscale() * (sqrt(0.75) * _edge); //Use 3D pythagorean theorem.
			break;
		case OCTAHEDRON:
			_actualradius = maxscale();
			break;
		case DODECAHEDRON:
			scale(1.0/dodecradius);
			_actualradius = 1.0;
			break;
		case ICOSOHEDRON:
			_actualradius = maxscale();
			break;
		case SPHERE:
			_actualradius = maxscale();
			break;
		case CYLINDER:
			bigradius = __max(__max(_scale.x(), _scale.y()) * _bottomradius,
				__max(_scale.x(), _scale.y()) * _topradius);
			_actualradius = sqrt(0.25*_scale.z()*_scale.z()*_height*_height + 
				bigradius * bigradius);
			_offset = cVector(0.0, 0.0, - 0.5 * _height);
			break;
		case CONE:
			bigradius = __max(_scale.x(), _scale.y()) * _bottomradius;
			_actualradius = sqrt(0.25*_height*_height + bigradius * bigradius);
			_offset = cVector(0.0, 0.0, -(1.0 - pow(0.5, 0.3333))* _height);
				/* A cone has volume 0.333 * base area * height, and its
				center of mass will be (1.0 - cube root of a half)*height up
				from the base. */
			break;
		case TORUS:
			_actualradius = maxscale() * (_radius + _tuberadius);
			break;
		case TEAPOT:
			_actualradius = maxscale() * _radius;
			break;
	}
}

//------------------Copy and operator= -----------------
void cPolyhedron::copy(const cPolyhedron &cpoly)
{
	if (this == &cpoly)
		return;
/* Because our class has some vector fields, we can't use the default overloaded
copy constructor and operator=.  So as to avoid having to maintain similar code
for these two different methods, we write a helper copy function that both
the copy constructor and the operator= can use. */
	_gluttype = cpoly._gluttype;
	_actualradius = cpoly._actualradius;
	_radius = cpoly._radius;
	_edge = cpoly._edge;
	_tuberadius = cpoly._tuberadius;
	_bottomradius = cpoly._bottomradius;
	_topradius = cpoly._topradius;
	_height = cpoly._height;
	_slices = cpoly._slices;
	_stacks = cpoly._stacks;
	_filled = cpoly._filled;
	_color = cpoly._color;
	_alpha = cpoly._alpha;
	_linewidth = cpoly._linewidth;
	_offset = cpoly._offset;
	_scale = cpoly._scale;
	_spinaxis = cpoly._spinaxis;
	_spinangle = cpoly._spinangle;
	_position = cpoly._position; 
	_spinrate = cpoly._spinrate;
}

cPolyhedron cPolyhedron::operator=(const cPolyhedron &cpoly)
{
	if (this == &cpoly)
		return *this;
	copy(cpoly);
	return *this;
}

cPolyhedron::cPolyhedron(const cPolyhedron &cPolyhedron)
{
	copy(cPolyhedron);
}

cPolyhedron::cPolyhedron(cPolyhedron *ppolygon)
{
	copy(*ppolygon);
}

//-----------------Mutators-----------------

void cPolyhedron::scale(cVector newscale)
{
	_scale = cVector(_scale.x()*newscale.x(), _scale.y()*newscale.y(),
		_scale.z()*newscale.z());
}

void cPolyhedron::scale(Real s)
{
	cVector sv(s, s, s);
	scale(sv);
}

void cPolyhedron::randomize()
{
	_gluttype = cRandomizer::pinst()->random(TETRAHEDRON,TEAPOT);
		//Randomizer includes low and hi values of the argumetns. 
//	_gluttype = TORUS; //Can set if I just want to see one kind.
	fixActualRadius(); 
	_color = cRandomizer::pinst()->randomColor(0.5);  //Arg is brightness.
	_filled = cRandomizer::pinst()->randomBool(0.9);
}

//============Key Methods=================

void cPolyhedron::animateTumble(Real dt)
{
	_spinangle += dt*_spinrate;
}

void cPolyhedron::draw(int drawflags)
{
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
/* We make the OpenGL calls in the reverse order of how we want them to 
happen.  That is, the last call will act first on the vertices of the
polyhedron.  What we want to do is draw the polyhedron, possibly shift
it by an _offset, then scale it, then rotate it, then translate it to its
_position.  The _offset is presently used to position the cylinders and 
cones so their center of mass seems to be at the origin. */
	glTranslatef(_position.x(), _position.y(), _position.z());
	glRotatef(_spinangle, _spinaxis.x(), _spinaxis.y(), _spinaxis.z());
	glScalef(_scale.x(), _scale.y(), _scale.z());
	glTranslatef(_offset.x(), _offset.y(), _offset.z());
	setMaterialColor(_color);
	if (filled() && drawflags == 0)
	{
		switch(_gluttype)
		{
			case TETRAHEDRON:
				glutSolidTetrahedron();
				break;
			case CUBE:
				glutSolidCube(_edge);
					//argument is value of cube edge.
				break;
			case OCTAHEDRON:
				glutSolidOctahedron();
				break;
			case DODECAHEDRON:
				glutSolidDodecahedron();
				break;
			case ICOSOHEDRON:
				glutSolidIcosahedron();
				break;
			case SPHERE:
				glutSolidSphere(_radius, _slices, _stacks);
				break;
			case CYLINDER:
				glutSolidCylinder(_bottomradius, _topradius, _height,
					_slices, _stacks);
				break;
			case CONE:
				glutSolidCone(_bottomradius, _height, _slices, _stacks);
				break;
			case TORUS:
				glutSolidTorus(_tuberadius, _radius,
					_slices, _stacks);
				break;
			case TEAPOT:
				glutSolidTeapot(_radius);
				break;
		}
	}
	else //!filled() or you pass in a non-zero drawflag 
	{
		glLineWidth(_linewidth);
		switch(_gluttype)
		{
			case TETRAHEDRON:
				glutWireTetrahedron();
				break;
			case CUBE:
				glutWireCube(_edge);
					//argument is value of cube edge.
				break;
			case OCTAHEDRON:
				glutWireOctahedron();
				break;
			case DODECAHEDRON:
				glutWireDodecahedron();
				break;
			case ICOSOHEDRON:
				glutWireIcosahedron();
				break;
			case SPHERE:
				glutWireSphere(_radius, _slices, _stacks);
				break;
			case CYLINDER:
				glutWireCylinder(_bottomradius, _topradius, _height,
					_slices, _stacks);
				break;
			case CONE:
				glutWireCone(_bottomradius, _height, _slices, _stacks);
				break;
			case TORUS:
				glutWireTorus(_tuberadius, _radius,
					_slices, _stacks);
				break;
			case TEAPOT:
				glutWireTeapot(_radius);
				break;
		}
	}
	::glPopMatrix();
#ifdef TESTRADIUSACCURACY
	::glPushMatrix();
	GLfloat testmaterialAmbDiff[4]={ 1.0f, 1.0f, 0.0f, 0.5f };
	glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, testmaterialAmbDiff );
	glTranslatef(_position.x(), _position.y(), _position.z());
	glRotatef(_spinangle, _spinaxis.x(), _spinaxis.y(), _spinaxis.z());
	glutWireSphere(1.1*_actualradius, 8, 8);
	::glPopMatrix();
#endif //TESTRADIUSACCURACY
}


