#ifndef POLYHEDRON_H
#define POLYHEDRON_H

#include "realnumber.h"
#include "vectormatrix.h"

//Helper methods
void setMaterialColor(cVector3 color); 
void setVertexColor(cVector3 color, float alpha = 1.0);
void glutWireCylinder(double bottomradius, double topradius, double height,
	int slices, int stacks);
void glutSolidCylinder(double bottomradius, double topradius, double height,
	int slices, int stacks);

class cPolyhedron 
{
public:
//const statics 
	static const int TETRAHEDRON;
	static const int CUBE;
	static const int OCTAHEDRON;
	static const int ICOSOHEDRON;
	static const int DODECAHEDRON;
	static const int SPHERE;
	static const int CYLINDER;
	static const int CONE;
	static const int TORUS;
	static const int TEAPOT;
protected:
//Type
	int _gluttype;
//Geometry
	Real _actualradius; /* Make this be the actual radius of the most
		distant point form the origin. */
	Real _radius; //For SPHERE, TORUS
	int _slices; //For SPHERE, TORUS, CYLINDER, CONE
	int _stacks; //For SPHERE, TORUS, CYLINDER, CONE
	Real _edge; //For CUBE
	Real _tuberadius; //For TORUS
	Real _bottomradius; //For CYLINDER and CONE
	Real _topradius; //For CYLINDER
	Real _height; //For CONE, CYLINDER
//Appearance
	cVector3 _color;
	Real _alpha;
	bool _filled;
	Real _linewidth;
//Transformation
	cVector _offset;  /* Correction displacement, if we always want
		to position the thing at certain offset from its position. */
	cVector _scale;
	Real _spinangle;
	cVector _spinaxis;
	cVector _position;
//Tumble
	Real _spinrate;
public:
//Constructor, destructor, operator=
	cPolyhedron(int gluttype = SPHERE, Real radius = 1.0,
		int slices = 32, int stacks = 16, Real edge = sqrt(4.0/3.0),
		Real tuberadius = 0.3,	Real bottomradius = 1.0,
		Real topradius = 1.0 , Real height = 3.0);
		/* I select the default edge value to make a cube of radius 1. */
	virtual	~cPolyhedron(){}; /* The destructor of a base class must be virtual
		if any of its child classes can have more fields that may need additional
		destruction. */
	void cPolyhedron::copy(const cPolyhedron &cpoly);
	cPolyhedron(const cPolyhedron &poly); //Uses copy
	cPolyhedron(cPolyhedron *ppoly); //Uses copy
	cPolyhedron operator=(const cPolyhedron &cpolygon); //uses the copy method
//Helper functions
	Real maxscale(){return __max(_scale.x(), __max(_scale.y(), _scale.z()));}
	void fixActualRadius(); /* When you construct a polyhedron you may need to fix
		the _scale to make it truly be the size that you expect.  In particular,
		we want to scale a default dodecahedron so it has unit radius like
		the other Platonic solids (except for cube, which takes an edge
		argument to its constructor). */
//Mutators
	//Color and draw style Mutators
	void setFilled(bool yesno){_filled = yesno;}
	void setColor(cVector3 col){_color = col;}
	void setLineWidth(Real linewidth){_linewidth = linewidth;}
	void setAlpha(Real alpha){_alpha = alpha;}
	//Geometric Mutators
	void randomize();
	void setPosition(const cVector &newposition){_position = newposition;}
	void scale(cVector scalevector);
	void scale(Real s);
	void translate(cVector dposition){_position += dposition;}
	//Shape Mutators
	void setGluttype(int gluttype){_gluttype = gluttype;}
	void setSlices(int slices){_slices = slices;}
	void setStacks(int stacks){_stacks = stacks;}
	void setEdge(Real edge){_edge = edge; fixActualRadius();}
	void setTubeRadius(Real tuberadius){_tuberadius = tuberadius;}
	void setBottomRadius(Real bottomradius)
		{_bottomradius = bottomradius; fixActualRadius();}
	void setTopRadius(Real topradius)
		{_topradius = topradius; fixActualRadius();}
	void setHeight(Real height){_height = height; fixActualRadius();}
//Accessors
	//Color and style accessors
	cVector color() const{return _color;}
	Real alpha()const{return _alpha;}
	bool filled() const{return _filled;}
	Real lineWidth() const {return _linewidth;}
	//Geometry accessors
	Real radius() const { return _actualradius * _scale.magnitude();}
	cVector position()const{return _position;}
	cVector offset(){return _offset;}
	//Shape accessors
	int glutType(){return _gluttype;}
//Key methods
	void animateTumble(Real dt);
	void draw(int drawflags = 0); //If drawflags is 1, we draw wireframe.
};
#endif //POLYHEDRON_H